// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <cstdlib>
#include <unistd.h>
#include <string>
#include <vector>
#include <iostream>

#include "options.h"
#include "cmdline.h"

// Constructor parses and builds options flags and argument vector.  There
// must be at least one argument after flags, and a few combinations of
// flags are invalid.  Exits on invalid command options or arguments.
CommandLine::CommandLine (int argc,
                          char * const * argv)
{
  program_name_ = argv[0];
  std::string::size_type p = program_name_.find_last_of ('/');
  if (p != program_name_.npos)
    program_name_.erase (0, p + 1);

  int c;
  while ((c = getopt (argc, argv, ":gpf:OPSsMvh")) != -1)
    {
      switch (c)
        {
        case 'g':
          options_.debugging_flag_ = true;
          break;
        case 'p':
          options_.profiling_flag_ = true;
          break;
        case 'f':
          if (std::string (optarg) != "PIC")
            {
              std::cerr << program_name_ << ": invalid option -f"
                << optarg << std::endl;
              usage ();
              std::exit (EXIT_FAILURE);
            }
          options_.PIC_flag_ = true;
          break;
        case 'O':
          options_.optimize_flag_ = true;
          break;
        case 'P':
          options_.intermediate_flag_ = true;
          break;
        case 's':
        case 'S':
          options_.assembly_flag_ = true;
          break;
        case 'M':
          options_.mangle_flag_ = true;
          break;
        case 'v':
          version ();
          std::exit (EXIT_SUCCESS);
        case 'h':
          help ();
          std::exit (EXIT_SUCCESS);
        case ':':
        case '?':
          std::cerr << program_name_ << ": invalid option -- "
            << static_cast<char>(optopt) << std::endl;
          usage ();
          std::exit (EXIT_FAILURE);
        default:
          usage ();
          std::exit (EXIT_FAILURE);
        }
    }

  if (optind >= argc)
    {
      std::cerr << program_name_
        << (options_.mangle_flag_ ? ": no ids given" : ": no input files")
        << std::endl;
      usage ();
      std::exit (EXIT_FAILURE);
    }

  if (options_.debugging_flag_ && options_.profiling_flag_)
    {
      std::cerr << program_name_
        << ": -g flag overrides -p flag; -p ignored" << std::endl;
      options_.profiling_flag_ = false;
    }
  if (options_.mangle_flag_
      && (options_.debugging_flag_ || options_.profiling_flag_
           || options_.PIC_flag_ || options_.optimize_flag_
           || options_.intermediate_flag_ || options_.assembly_flag_))
    {
      std::cerr << program_name_
        << ": -M cannot be used with other flags; -M ignored" << std::endl;
      options_.mangle_flag_ = false;
    }

  for (int i = optind; i < argc; i++)
    arguments_.push_back (std::string (argv[i]));
}

// Informational helpers.
void
CommandLine::help ()
{
  std::cerr << "Usage: " << program_name_
            << " [ options ] file [ file ... ]" << std::endl
            << "       " << program_name_
            << " -M id [ id ... ]" << std::endl << std::endl
    << "Options:" << std::endl
    << " -g      Include debugging information in the output files"
    << std::endl
    << " -p      Include line profiling for gmon.out execution profile"
    << std::endl
    << " -fPIC   Generate position independent code, for shared libraries"
    << std::endl
    << " -O      Optimize generated code (minimal optimization only)"
    << std::endl
    << " -P      Write out intermediate file (.p) during compilation"
    << std::endl
    << " -S,-s   Write out assembly language file (.s) during compilation"
    << std::endl
    << " -M      Name-mangle each id given, for linking with 'C'"
    << std::endl
    << " -v      Print compiler version number and exit"
    << std::endl
    << " -h      Print this help message"
    << std::endl;
}

void
CommandLine::version ()
{
  std::cout << std::endl << program_name_ << " version 1.4 ("
    << __DATE__ << ' ' << __TIME__ << ')' << std::endl
    << "Copyright (C) 2005 Simon Baldwin" << std::endl << std::endl
    << "This program comes with ABSOLUTELY NO WARRANTY; for details"
    << std::endl
    << "please see the file 'COPYING' supplied with the source code."
    << std::endl
    << "This is free software, and you are welcome to redistribute it"
    << std::endl
    << "under certain conditions; again, see 'COPYING' for details."
    << std::endl
    << "This program is released under the GNU General Public License."
    << std::endl;
}

void
CommandLine::usage ()
{
  std::cerr << "Try '" << program_name_
    << " -h' for more information." << std::endl;
}
