// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <unistd.h>
#include <cstdlib>
#include <climits>
#include <cctype>
#include <cassert>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <vector>
#include <map>
#include <algorithm>
#include <iterator>

#include "stack.h"
#include "register.h"
#include "options.h"
#include "program.h"
#include "dattable.h"
#include "data.h"
#include "symtable.h"
#include "symbol.h"
#include "optable.h"
#include "opcode.h"

// Code generator functions.  These belong to assorted classes, but are
// concentrated here to keep cpu-specific code in a single module.

// Stack and register definitions.
const Stack Stack::DataStack = Stack ("^D", "__fdpush", "__fdpop");
const Stack Stack::ReturnStack = Stack ("^R", "__frpush", "__frpop");
const Stack Stack::FloatStack = Stack ("^F", "__ffpush", "__ffpop");

const Register Register::R0 = Register ("R0", "%eax");
const Register Register::R1 = Register ("R1", "%edx");

// Constants for emitting debug.
namespace {
  const int CELL_SIZE = sizeof (int);
}
namespace Stabtype {
  enum { VOID = 1, CELL = 2, UCELL = 3, CHAR = 4 };
}

// Hack to suppress unused arguments warning from the compiler.
namespace {
  inline void
  args_unused (std::ostream & outs,
               const Options & options)
  {
    outs << (options.include_debugging () ? "" : "");
  }

  inline void
  args_unused (const Options & options)
  {
    options.include_debugging ();
  }
}

// Overall program generators.
void
Program::generate_preamble (std::ostream & outs) const
{
  outs << "\t.file\t" << '"' << source_path_ << '"' << std::endl
       << "\t.version\t" << '"' << "01.01" << '"' << std::endl
       << "forth_compiled.:" << std::endl;

  if (options_.include_profiling ())
    outs << "\tnop" << std::endl;

  // Debugger preamble identifies module, and sets up language variable types.
  if (options_.include_debugging ())
    {
      if (char * cwd = getcwd (NULL, 0)) // Linux extension to getcwd
        {
          outs << ".stabs\t" << '"' << cwd << '/' << '"' << ",100,0,0,Ltext0"
               << std::endl;
          std::free (cwd);
        }

      outs << ".stabs\t" << '"' << source_path_ << '"' << ",100,0,0,Ltext0"
           << std::endl
           << "\t.stabs\t" << '"' << "forth_compiled." << '"' << ",0x3c,0,0,0"
           << std::endl;

      outs << ".text" << std::endl
           << "Ltext0:" << std::endl;
      outs << ".stabs\t" << '"' << "void:t" << Stabtype::VOID << '='
           << Stabtype::VOID << ';' << '"' << ",128,0,0,0" << std::endl;
      outs << ".stabs\t" << '"' << "CELL:t" << Stabtype::CELL << '=' << 'r'
           << Stabtype::CELL << ';' << INT_MIN << ';' << INT_MAX << ';'
           << '"' << ",128,0,0,0" << std::endl;
      outs << ".stabs\t" << '"' << "UCELL:t" << Stabtype::UCELL << '=' << 'r'
           << Stabtype::CELL << ';' << 0 << ';'
           << static_cast<int>(UINT_MAX) << ';'
           << '"' << ",128,0,0,0" << std::endl;
      outs << ".stabs\t" << '"' << "CHAR:t" << Stabtype::CHAR << '=' << 'r'
           << Stabtype::CHAR << ';' << 0 << ';' << CHAR_MAX << ';'
           << '"' << ",128,0,0,0" << std::endl;
    }
}

void
Program::generate_postamble (std::ostream & outs) const
{
  if (options_.include_debugging ())
    {
      outs << "\t.text" << std::endl
           << "\t.stabs\t" << '"' << '"' << ",100,0,0,.Letext" << std::endl
           << ".Letext:" << std::endl;
    }

  outs << "\t.ident\t " << '"' << "Forthc: Forth 1.4, "
       << __DATE__ << ' ' << __TIME__ << ", Linux" << '"' << std::endl;
}

// Read-only data generators.
void
DataTable::generate (std::ostream & outs) const
{
  if (!vector_.empty ())
    outs << ".section\t.rodata" << std::endl;

  std::vector<std::vector<int> > synonyms (vector_.size ());

  // Create a vector of synonyms first; on writing, labels are duplicated
  // for synonyms, and data with synonyms is suppressed.
  for (unsigned int i = 0; i < vector_.size (); ++i)
    {
      if (vector_[i]->is_synonym ())
        synonyms[vector_[i]->get_synonym ()].push_back (vector_[i]->get_id ());
    }

  // Write with synonmyms suppressed.
  for (unsigned int i = 0; i < vector_.size (); ++i)
    {
      if (!vector_[i]->is_synonym ())
        {
          outs << ".LC" << i << ':' << std::endl;

          for (unsigned int j = 0; j < synonyms[i].size (); ++j)
            outs << ".LC" << (synonyms[i])[j] << ':' << std::endl;

          vector_[i]->generate (outs);
        }
    }
}

namespace {
  std::string
  data_to_string (const std::vector<unsigned char> & chunk)
  {
      std::ostringstream outs;
      outs.fill ('0');

      for (unsigned int i = 0; i < chunk.size (); ++i)
        {
          if (chunk[i] && isprint (chunk[i]))
            outs << chunk[i];
          else
            {
              outs << '\\' << std::oct << std::setw (3)
                   << static_cast<int>(chunk[i])
                   << std::setw (0) << std::dec;
            }
        }

      return outs.str ();
  }
}

void
Data::generate (std::ostream & outs) const
{
  const std::vector<unsigned char> & chunk = get_chunk ();

  outs << "\t.string " << '"'
       << data_to_string (chunk) << '"' << std::endl;
}

void
CountedStringData::generate (std::ostream & outs) const
{
  const std::vector<unsigned char> & chunk = get_chunk ();

  std::vector<unsigned char> size;
  size.push_back (static_cast<unsigned char>(chunk.size ()));

  outs << "\t.string " << '"' << data_to_string (size)
       << data_to_string (chunk) << '"' << std::endl;
}

// Variables and constants generators.
void
SymbolTable::generate (std::ostream & outs,
                       const Options & options) const
{
  for (SymbolTableMapIterator iter = map_.begin ();
      iter != map_.end (); ++iter)
    {
      iter->second->generate (outs, options);
    }
}

void
VariableSymbol::generate (std::ostream & outs,
                          const Options & options) const
{
  if (options.include_debugging ())
    {
      outs << ".stabs\t" << '"' << get_name ();

      if (units_ == CELL && size_ == 1)
        outs << ":G" << Stabtype::CELL;
      else
        {
          outs << ":Gar" << Stabtype::CELL << ";0;" << size_ - 1 << ';'
               << ((units_ == CELL) ? Stabtype::CELL : Stabtype::CHAR);
        }

      outs << '"' << ",40,0,0," << get_name () << std::endl;
    }

  int bytes = (units_ == CELL) ? size_ * CELL_SIZE : size_;
  int alignment = (bytes >= 32) ? 32 : ((units_ == CELL) ? 4 : 1);

  outs << "\t.comm  " << get_name () << ',' << bytes << ','
       << alignment << std::endl;
}

void
ConstantSymbol::generate (std::ostream & outs,
                          const Options & options) const
{
  if (options.include_debugging ())
    {
      outs << ".stabs\t" << '"' << get_name () << ":G"
           << Stabtype::CELL << '"'
           << ",40,0,0," << get_name () << std::endl;
    }

  outs << "\t.comm  " << get_name () << ',' << CELL_SIZE << ",4" << std::endl;
}

void
ExternalSymbol::generate (std::ostream & outs,
                          const Options & options) const
{
  // Special case; external _argc, _argv, and _envp are treated as if
  // variables; they're also a little special in the parser.
  const std::string name = get_name ();
  if (name == "_argc" || name == "_argv" || name == "_envp")
    {
      if (options.include_debugging ())
        {
          outs << ".stabs\t" << '"' << get_name () << ":G"
               << Stabtype::CELL << '"'
               << ",40,0,0," << get_name () << std::endl;
        }

      outs << "\t.comm  " << get_name () << ',' << CELL_SIZE
           << ",4" << std::endl;
    }
}

void
Symbol::generate (std::ostream & outs,
                  const Options & options) const
{
  args_unused (outs, options);
}

// Opcode generators.
namespace {
  const Symbol * current_definition = 0;
}

void
OpcodeTable::generate (std::ostream & outs,
                       const Options & options,
                       const SymbolTable & symtable) const
{
  // Split opcodes into two sets; those inside definitions, and those not.
  std::vector<const Opcode *> opcodes, main_opcodes;
  bool in_main = true;

  for (unsigned int i = 0; i < vector_.size (); ++i)
    {
      if (vector_[i]->is_define_opcode ())
        in_main = false;

      std::vector<const Opcode *> & table = in_main ? main_opcodes : opcodes;
      table.push_back (vector_[i]);

      if (vector_[i]->is_enddefine_opcode ())
        in_main = true;
    }

  // Combine sets, appending opcodes not in definitions into a synthesized
  // main definition.  We expect the parser to have added the "main" symbol.
  DefineOpcode * main_start = 0;
  LabelOpcode * main_label = 0;
  EndDefineOpcode * main_end = 0;

  if (!main_opcodes.empty ())
    {
      const DefinableSymbol * main =
                            symtable.lookup ("main")->is_definable_symbol ();
      assert (main);

      int main_first_line = main_opcodes[0]->get_line ();

      main_start = new DefineOpcode (main, main_first_line);
      opcodes.push_back (main_start);

      copy (main_opcodes.begin (),
            main_opcodes.end (), back_inserter (opcodes));

      int main_last_line = main_opcodes[main_opcodes.size () - 1]->get_line ();

      main_label = new LabelOpcode (Label (0), main_last_line);
      opcodes.push_back (main_label);

      main_end = new EndDefineOpcode (main, main_last_line);
      opcodes.push_back (main_end);
    }

  // Traverse combined opcodes, generating code for each.
  current_definition = 0;
  int current_line = -1;

  for (unsigned int i = 0; i < opcodes.size (); ++i)
    {
      const Opcode * opcode = opcodes[i];

      // Write opcode debugging data for line number if required.
      if (options.include_debugging ()
          && current_line != opcode->get_line ())
        {
          // Don't write debugger stabs for pseudo-opcodes or assembly.
          bool requires_debug = !(opcode->is_label_opcode ()
                                  || opcode->is_define_opcode ()
                                  || opcode->is_enddefine_opcode ()
                                  || opcode->is_assembly_opcode ());

          if (requires_debug)
            {
              assert (current_definition);
              outs << ".stabn 68,0," << opcode->get_line () << ",.LM"
                   << opcode->get_id () << "-"
                   << current_definition->get_name () << std::endl
                   << ".LM" << opcode->get_id () << ':' << std::endl;
            }
        }

      // Generate opcode assembly.
      opcode->generate (outs, options);

      current_line = opcode->get_line ();
    }

  delete main_start;
  delete main_label;
  delete main_end;
}

void
CallOpcode::generate (std::ostream & outs,
                      const Options & options) const
{
  outs << "\tcall " << symbol_->get_name ()
       << (options.position_independent () ? "@PLT" : "") << std::endl;
}

void
DefineOpcode::generate (std::ostream & outs,
                        const Options & options) const
{
  // Function preamble.
  outs << ".text" << std::endl
       << "\t.align 4" << std::endl;
  
  std::string symbol_name = symbol_->get_name ();

  if (options.include_debugging ())
    {
      outs << ".stabs " << '"' << symbol_name << ":F" << Stabtype::VOID << '"'
           << ",36,0,0," << symbol_name << std::endl;
    }

  outs << ".globl " << symbol_name << std::endl
       << "\t.type\t" << symbol_name << ",@function" << std::endl
       << symbol_name << ':' << std::endl;

  if (options.include_debugging ())
    {
      outs << ".stabn 68,0," << get_line () << ",.LMa"
           << get_id () << "-" << symbol_name << std::endl
           << ".LMa" << get_id () << ':' << std::endl;
    }

  outs << "\tpushl %ebp" << std::endl
       << "\tmovl %esp,%ebp" << std::endl;

  // Preserve registers.  Careful reading of the Intel ABI reveals that it's
  // necessary only to preserve selected registers, and as we don't even use
  // all of them, just the following few suffice.
  if (symbol_->is_codeword_symbol ())
    {
      outs << "\tpushl %edi" << std::endl
           << "\tpushl %esi" << std::endl
           << "\tpushl %ebx" << std::endl;
    }
  else if (options.position_independent ())
    outs << "\tpushl %ebx" << std::endl;

  int symbol_id = symbol_->get_id ();

  if (options.position_independent ())
    {
      outs << "\tcall .Lpic" << symbol_id << std::endl
           << ".Lpic" << symbol_id << ':' << std::endl
           << "\tpopl %ebx" << std::endl
           << "\taddl $_GLOBAL_OFFSET_TABLE_"
           << "+[.-.Lpic" << symbol_id << "],%ebx" << std::endl;

      // Write a label usable for PIC compilation flag checking.
      outs << "0:" << std::endl;
    }

  // Set up argc, argv, and envp for main function.
  if (symbol_name == "main")
    {
      outs << "\tmovl 0x08(%ebp),%eax" << std::endl;
      if (options.position_independent ())
        {
          outs << "\tmovl _argc@GOT(%ebx),%ecx" << std::endl
               << "\tmovl %eax,(%ecx)" << std::endl;
        }
      else
        outs << "\tmovl %eax,_argc" << std::endl;

      outs << "\tmovl 0x0C(%ebp),%eax" << std::endl;
      if (options.position_independent ())
        {
          outs << "\tmovl _argv@GOT(%ebx),%ecx" << std::endl
               << "\tmovl %eax,(%ecx)" << std::endl;
        }
      else
        outs << "\tmovl %eax,_argv" << std::endl;

      outs << "\tmovl 0x10(%ebp),%eax" << std::endl;
      if (options.position_independent ())
        {
          outs << "\tmovl _envp@GOT(%ebx),%ecx" << std::endl
               << "\tmovl %eax,(%ecx)" << std::endl;
        }
      else
        outs << "\tmovl %eax,_envp" << std::endl;
    }

  if (options.include_profiling ())
    {
      outs << ".data" << std::endl
           << "\t.align 4" << std::endl
           << ".LP" << symbol_id << ':' << std::endl
           << "\t.long 0" << std::endl
           << ".text" << std::endl;

      if (options.position_independent ())
        {
          outs << "\tleal .LP" << symbol_id << "@GOTOFF(%ebx),%edx"
               << std::endl
               << "\tcall *mcount@GOT(%ebx)" << std::endl;
        }
      else
        {
          outs << "\tmovl $.LP" << symbol_id << ",%edx" << std::endl
               << "\tcall mcount" << std::endl;
        }
    }

  if (options.include_debugging ())
    {
      outs << ".stabn 68,0," << get_line () << ",.LMb"
           << get_id () << "-" << symbol_name << std::endl
           << ".LMb" << get_id () << ':' << std::endl;

      outs << ".LBB" << symbol_id << ':' << std::endl;
    }

  if (symbol_->is_codeword_symbol ())
      outs << "#APP" << std::endl;

  current_definition = symbol_;
}

void
EndDefineOpcode::generate (std::ostream & outs,
                           const Options & options) const
{
  if (symbol_->is_codeword_symbol ())
      outs << "#NO_APP" << std::endl;

  std::string symbol_name = symbol_->get_name ();
  int symbol_id = symbol_->get_id ();

  if (options.include_debugging ())
    {
      outs << ".LBE" << symbol_id << ':' << std::endl;

      outs << ".stabn 68,0," << get_line () << ",.LM"
           << get_id () << "-" << symbol_name << std::endl
           << ".LM" << get_id () << ':' << std::endl;
    }

  // Restore registers.
  if (symbol_->is_codeword_symbol ())
    {
      outs << "\tpopl %ebx" << std::endl
           << "\tpopl %esi" << std::endl
           << "\tpopl %edi" << std::endl;
    }
  else if (options.position_independent ())
      outs << "\tpopl %ebx" << std::endl;

  // Function postamble.
  outs << "\tleave" << std::endl
       << "\tret" << std::endl;

  outs << ".Lfe" << symbol_id << ':' << std::endl
       << "\t.size\t " << symbol_name << ",.Lfe" << symbol_id << '-'
       << symbol_name << std::endl;

  if (options.include_debugging ())
    {
      outs << ".stabn\t192,0,0,.LBB" << symbol_id << '-'
           << symbol_name << std::endl
           << ".stabn\t224,0,0,.LBE" << symbol_id << '-'
           << symbol_name << std::endl
           << ".Lscope" << symbol_id << ':' << std::endl
           << ".stabs\t" << '"' << '"' << ",36,0,0,"
           << ".Lscope" << symbol_id << '-' << symbol_name << std::endl;
    }

  current_definition = 0;
}

void
LabelOpcode::generate (std::ostream & outs,
                       const Options & options) const
{
  args_unused (options);
  outs << ".L" << label_.get_value () << ':' << std::endl;
}

void
AssemblyOpcode::generate (std::ostream & outs,
                          const Options & options) const
{
  args_unused (options);
  outs << assembly_ << std::endl;
}

void
LoadValueOpcode::generate (std::ostream & outs,
                           const Options & options) const
{
  int value = value_.get_value ();

  if (options.optimize_code () && value ==  0)
    {
      outs << "\txorl " << reg_.get_cpu_name () << ','
           << reg_.get_cpu_name () << std::endl;
    }
  else
    outs << "\tmovl $" << value << ',' << reg_.get_cpu_name () << std::endl;
}

void
LoadRegisterOpcode::generate (std::ostream & outs,
                              const Options & options) const
{
  args_unused (options);
  outs << "\tmovl " << with_.get_cpu_name () << ','
       << reg_.get_cpu_name () << std::endl;
}

void
LoadSymbolOpcode::generate (std::ostream & outs,
                            const Options & options) const
{
  if (options.position_independent ())
    {
      outs << "\tmovl " << symbol_->get_name () << "@GOT(%ebx),"
           << reg_.get_cpu_name () << std::endl;
    }
  else
    {
      outs << "\tleal " << symbol_->get_name () << ','
           << reg_.get_cpu_name () << std::endl;
    }
}

void
LoadSymbolIndirectOpcode::generate (std::ostream & outs,
                                    const Options & options) const
{
  if (options.position_independent ())
    {
      outs << "\tmovl " << symbol_->get_name () << "@GOT(%ebx),%ecx"
           << std::endl
           << "\tmovl (%ecx)," << reg_.get_cpu_name () << std::endl;
    }
  else
    {
      outs << "\tmovl (" << symbol_->get_name () << "),"
           << reg_.get_cpu_name () << std::endl;
    }
}

void
LoadDataOpcode::generate (std::ostream & outs,
                          const Options & options) const
{
  if (options.position_independent ())
    {
      outs << "\tleal .LC" << data_->get_id () << "@GOTOFF(%ebx),"
           << reg_.get_cpu_name () << std::endl;
    }
  else
    {
      outs << "\tleal .LC" << data_->get_id () << ','
           << reg_.get_cpu_name () << std::endl;
    }
}

void
StoreSymbolIndirectOpcode::generate (std::ostream & outs,
                                     const Options & options) const
{
  if (options.position_independent ())
    {
      outs << "\tmovl " << symbol_->get_name () << "@GOT(%ebx),%ecx"
           << std::endl
           << "\tmovl " << reg_.get_cpu_name () << ",(%ecx)" << std::endl;
    }
  else
    {
      outs << "\tmovl " << reg_.get_cpu_name () << ",("
           << symbol_->get_name () << ')' << std::endl;
    }
}

void
PushOpcode::generate (std::ostream & outs,
                      const Options & options) const
{
  bool need_exchange = !(reg_.get_cpu_name () == "%eax");

  if (need_exchange)
    outs << "\txchgl " << reg_.get_cpu_name () << ",%eax" << std::endl;

  outs << "\tcall " << stack_.get_push_function ()
       << (options.position_independent () ? "@PLT" : "") << std::endl;

  if (need_exchange)
    outs << "\txchgl %eax," << reg_.get_cpu_name () << std::endl;
}

void
PopOpcode::generate (std::ostream & outs,
                     const Options & options) const
{
  bool need_exchange = !(reg_.get_cpu_name () == "%eax");

  if (need_exchange)
    outs << "\txchgl " << reg_.get_cpu_name () << ",%eax" << std::endl;

  outs << "\tcall " << stack_.get_pop_function ()
       << (options.position_independent () ? "@PLT" : "") << std::endl;

  if (need_exchange)
    outs << "\txchgl %eax," << reg_.get_cpu_name () << std::endl;
}

void
AddValueOpcode::generate (std::ostream & outs,
                          const Options & options) const
{
  int value = value_.get_value ();

  if (options.optimize_code ()
      && (value == 1 || value == -1 || value ==  0))
    {
      switch (value)
        {
        case  1:
          outs << "\tincl " << reg_.get_cpu_name () << std::endl;
          break;
        case -1:
          outs << "\tdecl " << reg_.get_cpu_name () << std::endl;
          break;
        case  0:
          break;
        }
    }
  else
    outs << "\taddl $" << value << ',' << reg_.get_cpu_name () << std::endl;
}

void
AddRegisterOpcode::generate (std::ostream & outs,
                             const Options & options) const
{
  args_unused (options);
  outs << "\taddl " << increment_.get_cpu_name () << ','
       << reg_.get_cpu_name () << std::endl;
}

void
SubtractValueOpcode::generate (std::ostream & outs,
                               const Options & options) const
{
  int value = value_.get_value ();

  if (options.optimize_code ()
      && (value == 1 || value == -1 || value ==  0))
    {
      switch (value)
        {
        case  1:
          outs << "\tdecl " << reg_.get_cpu_name () << std::endl;
          break;
        case -1:
          outs << "\tincl " << reg_.get_cpu_name () << std::endl;
          break;
        case  0:
          break;
        }
    }
  else
    outs << "\tsubl $" << value << ',' << reg_.get_cpu_name () << std::endl;
}

void
SubtractRegisterOpcode::generate (std::ostream & outs,
                                  const Options & options) const
{
  args_unused (options);
  outs << "\tsubl " << decrement_.get_cpu_name () << ','
       << reg_.get_cpu_name () << std::endl;
}

void
JumpOpcode::generate (std::ostream & outs,
                      const Options & options) const
{
  args_unused (options);
  outs << "\tjmp .L" << label_.get_value () << std::endl;
}

void
JumpZeroOpcode::generate (std::ostream & outs,
                          const Options & options) const
{
  args_unused (options);
  outs << "\ttestl " << reg_.get_cpu_name () << ','
       << reg_.get_cpu_name () << std::endl;
  outs << "\tje .L" << label_.get_value () << std::endl;
}

void
JumpNonZeroOpcode::generate (std::ostream & outs,
                             const Options & options) const
{
  args_unused (options);
  outs << "\ttestl " << reg_.get_cpu_name () << ','
       << reg_.get_cpu_name () << std::endl;
  outs << "\tjne .L" << label_.get_value () << std::endl;
}

void
JumpEqualOpcode::generate (std::ostream & outs,
                           const Options & options) const
{
  args_unused (options);
  outs << "\tcmpl " << reg1_.get_cpu_name () << ','
       << reg2_.get_cpu_name () << std::endl;
  outs << "\tje .L" << label_.get_value () << std::endl;
}

void
JumpNotEqualOpcode::generate (std::ostream & outs,
                              const Options & options) const
{
  args_unused (options);
  outs << "\tcmpl " << reg1_.get_cpu_name () << ','
       << reg2_.get_cpu_name () << std::endl;
  outs << "\tjne .L" << label_.get_value () << std::endl;
}

void
JumpGreaterEqualOpcode::generate (std::ostream & outs,
                                  const Options & options) const
{
  args_unused (options);
  outs << "\ttestl " << reg_.get_cpu_name () << ','
       << reg_.get_cpu_name () << std::endl;
  outs << "\tjge .L" << label_.get_value () << std::endl;
}

void
JumpLessEqualOpcode::generate (std::ostream & outs,
                               const Options & options) const
{
  args_unused (options);
  outs << "\ttestl " << reg_.get_cpu_name () << ','
       << reg_.get_cpu_name () << std::endl;
  outs << "\tjle .L" << label_.get_value () << std::endl;
}

void
NoOpOpcode::generate (std::ostream & outs,
                      const Options & options) const
{
  args_unused (outs, options);
}
