// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#ifndef _VNPFORTH_DATA_H
#define _VNPFORTH_DATA_H

#include <vector>
#include <string>

class DataTable;

// Abstract data class, holds untyped binary data and some housekeeping,
// including a sequential id for each object created.
class Data
{
public:
  virtual ~Data ();

  bool is_redefinition_of (const Data * data) const;
  void set_synonym_of (const Data * data);
  int get_id () const;

  bool is_synonym () const;
  int get_synonym () const;

  static const std::string create_list_header ();
  const std::string create_list_entry () const;
  virtual const std::string create_list_data () const = 0;

  virtual void generate (std::ostream & outs) const;

protected:
  Data (size_t size,
        void const * pod,
        int line,
        const std::string & type,
        DataTable & datatable);

  const std::vector<unsigned char> & get_chunk () const;
  const std::string get_type () const;

private:
  static int count;
  static int instances;

  const int id_;
  std::vector<unsigned char> chunk_;
  const int line_;
  const std::string type_;
  int synonym_of_;
};

// Inline function definitions.
inline bool
Data::is_redefinition_of (const Data * data) const
{
  return chunk_ == data->chunk_;
}

inline void
Data::set_synonym_of (const Data * data)
{
  synonym_of_ = data->id_;
}

inline int
Data::get_id () const
{
  return id_;
}

inline bool
Data::is_synonym () const
{
  return synonym_of_ != -1;
}

inline int
Data::get_synonym () const
{
  return synonym_of_;
}

inline const std::string
Data::get_type () const
{
  return type_;
}

inline const std::vector<unsigned char> &
Data::get_chunk () const
{
  return chunk_;
}

// Concrete plain binary data class.
class BinaryData:public Data
{
public:
  BinaryData (size_t size,
              void const * pod,
              int line,
              DataTable & datatable):
    Data (size, pod, line, "Binary", datatable)
  {
  }

private:
  const std::string create_list_data () const;
};

// Abstract text data class, and concrete string data classes.
class TextData:public Data
{
public:
  TextData (size_t size,
            void const * pod,
            int line,
            const std::string & type,
            DataTable & datatable):
    Data (size, pod, line, type, datatable)
  {
  }

private:
  const std::string create_list_data () const;
};

class CountedStringData:public TextData
{
public:
  CountedStringData (size_t size,
                     void const * pod,
                     int line,
                     DataTable & datatable):
    TextData (size, pod, line, "Cstring", datatable)
  {
  }

  void generate (std::ostream & outs) const;
};

class StringData:public TextData
{
public:
  StringData (size_t size,
              void const * pod,
              int line,
              DataTable & datatable):
    TextData (size, pod, line, "String", datatable)
  {
  }
};

#endif
