// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <string>
#include <algorithm>
#include <functional>

#include "data.h"
#include "dattable.h"

// Delete all data objects in the table, and destructor.
namespace {
  void
  dispose (const Data * data)
  {
    delete data;
  }
}
void
DataTable::clear ()
{
  std::for_each (vector_.begin (), vector_.end (), dispose);
  vector_.erase (vector_.begin (), vector_.end ());
}
DataTable::~DataTable ()
{
  clear ();
}

// Add data objects, checking for duplicates and finding any synonyms.
namespace {
  bool
  duplicate_of (const Data * data,
                const Data * other)
  {
    return data->get_id () == other->get_id ();
  }
  bool
  redefinition_of (const Data * data,
                   const Data * other)
  {
    return data->is_redefinition_of (other);
  }
}
void
DataTable::add (Data * data)
{
  DataTableVectorIterator iter;
  iter = std::find_if (vector_.begin (), vector_.end (),
                       std::bind2nd (std::ptr_fun (duplicate_of), data));
  if (iter != vector_.end ())
    throw std::string ("DataTable::add -- duplicate table entry");

  iter = std::find_if (vector_.begin (), vector_.end (),
                       std::bind2nd (std::ptr_fun (redefinition_of), data));
  if (iter != vector_.end ())
    data->set_synonym_of (*iter);

  vector_.push_back (data);
}

// Pretty-print to create the symbol table listing.
namespace {
  void
  print_out (const Data * data,
             std::ostream * outs)
  {
    *outs << data->create_list_entry ();
  }
}
void
DataTable::create_listing (std::ostream & outs) const
{
  outs << "Data table:" << std::endl;
  if (vector_.empty ())
    {
      outs << " (empty)" << std::endl;
      return;
    }

  outs << Data::create_list_header ();
  std::for_each (vector_.begin (), vector_.end (),
                 std::bind2nd (std::ptr_fun (print_out), &outs));
}
