%{
// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005 Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//

#include <iostream>

#include "parser.h"

#define YYSTYPE char *
#define YYERROR_VERBOSE 1
extern int yylex ();

// External definition of active parser for non-OO yacc/lex code.
extern Parser * _parser;

// Macro to cause yyparse to abort if the maximum number of allowable errors
// has been exceeded.  This is an icky thing to have to do, but the closed
// nature of yyparse seems to leave little option - even a setjmp/longjmp is
// not an improvement.  And it doesn't even trigger until we hit a valid
// token, so we may well have a lot more errors than the maximum allowed by
// the time that happens.  All in all, rather unpleasant.
#define MNE do {                                          \
      if (_parser->excessive_errors ())                   \
        {                                                 \
          _parser->parse_error ()                         \
            << "too many errors, giving up" << std::endl; \
          YYABORT;                                        \
        }                                                 \
      } while (0)

// yyerror function links to the parser's error reporting.
namespace {
  void yyerror (const char * message)
  {
    _parser->parse_error () << message << std::endl;
  }
}

%}

%token VARIABLE CONSTANT
%token CREATE ALLOT
%token EXIT
%token IF THEN ELSE
%token DO QUERY_DO "?DO" LOOP PLUS_LOOP "+LOOP" LEAVE
%token _BEGIN "BEGIN" UNTIL WHILE REPEAT AGAIN
%token CASE OF ENDOF ENDCASE
%token RECURSE
%token PSTRING "print string"
%token CSTRING "counted string"
%token INTEGER_LITERAL "integer"
%token STRING_LITERAL "literal string"
%token FLOAT_LITERAL "float"
%token CHAR_LITERAL "char literal"
%token OBJECT_ADDRESS "address"
%token WORD "word"
%token CODE ASM "assembler string" END_CODE "END-CODE"

%%
program:        elements
                ;

elements:
                | elements element
                ;

element:          declaration
                | definition
                | statement
                | error WORD { MNE; yyerrok; }
                ;

declaration:      VARIABLE WORD { MNE; _parser->f_definedvariable ($2); }
                | CONSTANT WORD { MNE; _parser->f_definedconstant ($2); }
                | CREATE WORD INTEGER_LITERAL WORD ALLOT { MNE;
                                         _parser->f_definedarray ($2, $3, $4); }
                | CREATE WORD INTEGER_LITERAL ALLOT { MNE;
                                         _parser->f_definedarray ($2, $3); }
                ;

definition:       ':' definedword statements _semicolon
                | CODE codeword codes _end_code
                ;
definedword:      WORD { MNE; _parser->f_definedword ($1); }
                ;
_semicolon:       ';' { MNE; _parser->f_semicolon (); }
                ;
codeword:         WORD { MNE; _parser->f_codeword ($1); }
                ;
codes:            code
                | codes code
                ;
code:             ASM { MNE; _parser->f_code ($1); }
                ;
_end_code:        END_CODE { MNE; _parser->f_end_code (); }
                ;

statement:        if_statement
                | do_loop
                | do_leave
                | begin_loop
                | begin_while
                | case_statement
                | recurse
                | EXIT { MNE; _parser->f_exit (); }
                | PSTRING { MNE; _parser->f_pstring ($1); }
                | CSTRING { MNE; _parser->f_cstring ($1); }
                | INTEGER_LITERAL { MNE; _parser->f_integer_literal ($1); }
                | FLOAT_LITERAL { MNE; _parser->f_float_literal ($1); }
                | STRING_LITERAL { MNE; _parser->f_string_literal ($1); }
                | CHAR_LITERAL { MNE; _parser->f_char_literal ($1); }
                | OBJECT_ADDRESS { MNE; _parser->f_object_address ($1); }
                | ALLOT { MNE; _parser->f_word ($1); }
                | WORD  { MNE; _parser->f_word ($1); }
                ;

if_statement:     _if statements _then
                | _if statements _else statements _then
                ;
_if:              IF { MNE; _parser->f_if (); }
                ;
_then:            THEN { MNE; _parser->f_then (); }
                ;
_else:            ELSE { MNE; _parser->f_else (); }
                ;

do_loop:          _do statements _loop
                | _do statements _plus_loop
                ;
_do:              DO { MNE; _parser->f_do (); }
                | QUERY_DO { MNE; _parser->f_query_do (); }
                ;
_loop:            LOOP { MNE; _parser->f_loop (); }
                ;
_plus_loop:       PLUS_LOOP { MNE; _parser->f_plus_loop (); }
                ;
do_leave:         LEAVE { MNE; _parser->f_leave (); }
                ;

begin_loop:       _begin statements _until
                | _begin statements _again
                | _begin statements _repeat
                ;
_begin:           _BEGIN { MNE; _parser->f_begin (); }
                ;
_until:           UNTIL { MNE; _parser->f_until (); }
                ;
_again:           AGAIN { MNE; _parser->f_again (); }
                ;
_repeat:          REPEAT { MNE; _parser->f_repeat (); }
                ;
begin_while:      WHILE { MNE; _parser->f_while (); }
                ;

case_statement: _case casebody statements _endcase
                ;
casebody:         caseclause
                | casebody caseclause
                ;
caseclause:       statements _of statements _endof
                ;
_case:            CASE { MNE; _parser->f_case (); }
                ;
_endcase:         ENDCASE { MNE; _parser->f_endcase (); }
                ;
_of:              OF { MNE; _parser->f_of (); }
                ;
_endof:           ENDOF { MNE; _parser->f_endof (); }
                ;

recurse:          RECURSE { MNE; _parser->f_recurse (); }
                ;

statements:       elements
                ;
%%
