// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <string>
#include <sstream>
#include <iomanip>

#include "util.h"
#include "data.h"
#include "optable.h"
#include "opcode.h"

// Static opcode id sequence and instance counter.
int Opcode::count = 0;
int Opcode::instances = 0;

// Opcode constructor, auto-adds to the given opcode table, if any, and
// destructor, deletes any opcode optimized by this one.
Opcode::Opcode (int line,
                OpcodeTable * optable):
  id_ (count++), line_ (line), optimizes_ (0)
{
  instances++;
  if (optable)
    optable->add (this);
}

Opcode::~Opcode ()
{
  delete optimizes_;
  if (--instances == 0)
    count = 0;
}

// General and front-end opcode pretty printing function.
const std::string
Opcode::create_list_header ()
{
  std::ostringstream outs, outs2;

  outs << std::setw (8) << "Line" << std::setw (5) << "Id" << ' '
       << leftjustify << std::setw (30) << "Operation"
       << rightjustify << std::setw (0) << "Optimizes" << std::endl;

  outs2<< std::setw (8) << "-------" << std::setw (5) << "----" << ' '
       << leftjustify << std::setw (30) << "---------"
       << rightjustify << std::setw (0) << "---------" << std::endl;

  return outs.str () + outs2.str ();
}

const std::string
Opcode::create_list_entry () const
{
  std::ostringstream outs;

  outs << std::setw (8) << line_ << std::setw (5) << id_;
  if (optimizes_)
    {
      outs << ' ' << leftjustify << std::setw (30)
           << create_list_specific ()
           << rightjustify << optimizes_->create_list_specific ();
    }
  else
      outs << ' ' << std::setw (0)
           << create_list_specific ();
  outs << std::endl;

  return outs.str ();
}

// Pretty printing functions for specific opcodes.
const std::string
CallOpcode::create_list_specific () const
{
  return std::string (" call &") + to_string (symbol_->get_id ());
}

const std::string
DefineOpcode::create_list_specific () const
{
  return std::string ("function &") + to_string (symbol_->get_id ()) + " {";
}

const std::string
EndDefineOpcode::create_list_specific () const
{
  return std::string ("}");
}

const std::string
LabelOpcode::create_list_specific () const
{
  return std::string (" label %") + to_string (label_.get_value ());
}

const std::string
AssemblyOpcode::create_list_specific () const
{
  return std::string (" asm (") + assembly_ + ")";
}

const std::string
LoadValueOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- "
         + to_string (value_.get_value ());
}

const std::string
LoadRegisterOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- " + with_.get_name ();
}

const std::string
LoadSymbolOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- &"
         + to_string (symbol_->get_id ());
}

const std::string
LoadSymbolIndirectOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- @"
         + to_string (symbol_->get_id ());
}

const std::string
LoadDataOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- '"
         + to_string (data_->get_id ());
}

const std::string
StoreSymbolIndirectOpcode::create_list_specific () const
{
  return std::string (" &") + to_string (symbol_->get_id ())
         + " <- " + reg_.get_name ();
}

const std::string
PushOpcode::create_list_specific () const
{
  return std::string (" ") + stack_.get_name () + " <- " + reg_.get_name ();
}

const std::string
PopOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- " + stack_.get_name ();
}

const std::string
AddValueOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- " + reg_.get_name ()
         + " + " + to_string (value_.get_value ());
}

const std::string
AddRegisterOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- " + reg_.get_name ()
         + " + " + increment_.get_name ();
}

const std::string
SubtractValueOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- " + reg_.get_name ()
         + " - " + to_string (value_.get_value ());
}

const std::string
SubtractRegisterOpcode::create_list_specific () const
{
  return std::string (" ") + reg_.get_name () + " <- " + reg_.get_name ()
         + " - " + decrement_.get_name ();
}

const std::string
JumpOpcode::create_list_specific () const
{
  return std::string (" --> %") + to_string (label_.get_value ());
}

const std::string
JumpZeroOpcode::create_list_specific () const
{
  return std::string (" if ") + reg_.get_name () + "=0 --> %"
         + to_string (label_.get_value ());
}

const std::string
JumpNonZeroOpcode::create_list_specific () const
{
  return std::string (" if ") + reg_.get_name () + "!=0 --> %"
         + to_string (label_.get_value ());
}

const std::string
JumpEqualOpcode::create_list_specific () const
{
  return std::string (" if ") + reg1_.get_name () + "=" + reg2_.get_name ()
         + " --> %" + to_string (label_.get_value ());
}

const std::string
JumpNotEqualOpcode::create_list_specific () const
{
  return std::string (" if ") + reg1_.get_name () + "!=" + reg2_.get_name ()
         + " --> %" + to_string (label_.get_value ());
}

const std::string
JumpGreaterEqualOpcode::create_list_specific () const
{
  return std::string (" if ") + reg_.get_name () + ">=0 --> %"
         + to_string (label_.get_value ());
}

const std::string
JumpLessEqualOpcode::create_list_specific () const
{
  return std::string (" if ") + reg_.get_name () + "<=0 --> %"
         + to_string (label_.get_value ());
}

const std::string
NoOpOpcode::create_list_specific () const
{
  return std::string (" no-op");
}
