// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#ifndef _VNPFORTH_OPCODE_H
#define _VNPFORTH_OPCODE_H

#include <string>

#include "operand.h"
#include "register.h"
#include "stack.h"
#include "symbol.h"

class Data;
class OpcodeTable;
class Options;
class JumpOpcode;
class NoOpOpcode;
class LabelOpcode;
class PushOpcode;
class PopOpcode;
class DefineOpcode;
class EndDefineOpcode;
class AssemblyOpcode;

// Abstract opcode class, includes line number, and a pointer to any other
// opcode inserted to optimize this one away.
class Opcode
{
public:
  virtual ~Opcode ();
  static const std::string create_list_header ();
  const std::string create_list_entry () const;

  void set_optimizes (Opcode * optimizes);
  int get_id () const;
  int get_line () const;

  virtual void generate (std::ostream & outs,
                         const Options & options) const = 0;

  virtual const JumpOpcode * is_jump_opcode () const;
  virtual const NoOpOpcode * is_noop_opcode () const;
  virtual const LabelOpcode * is_label_opcode () const;
  virtual const PushOpcode * is_push_opcode () const;
  virtual const PopOpcode * is_pop_opcode () const;
  virtual const DefineOpcode * is_define_opcode () const;
  virtual const EndDefineOpcode * is_enddefine_opcode () const;
  virtual const AssemblyOpcode * is_assembly_opcode () const;

protected:
  Opcode (int line, OpcodeTable * optable = 0);

  virtual const std::string create_list_specific () const = 0;

private:
  static int count;
  static int instances;

  const int id_;
  const int line_;
  const Opcode * optimizes_;
};

// Inline function definitions.
inline void
Opcode::set_optimizes (Opcode * optimizes)
{
  delete optimizes_;
  optimizes_ = optimizes;
}

inline int
Opcode::get_id () const
{
  return id_;
}

inline int
Opcode::get_line () const
{
  return line_;
}

// Function definition and calls, and label opcode.
class CallOpcode:public Opcode
{
public:
  CallOpcode (const CallableSymbol * symbol,
              int line,
              OpcodeTable * optable = 0):
    Opcode (line, optable), symbol_ (symbol)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Symbol * symbol_;
};

class DefineOpcode:public Opcode
{
public:
  DefineOpcode (const DefinableSymbol * symbol,
                int line,
                OpcodeTable * optable = 0):
    Opcode (line, optable), symbol_ (symbol)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;
  const DefineOpcode * is_define_opcode () const;

protected:
  const std::string create_list_specific () const;
private:
  const Symbol * symbol_;
};

class EndDefineOpcode:public Opcode
{
public:
  EndDefineOpcode (const DefinableSymbol * symbol,
                   int line,
                   OpcodeTable * optable = 0):
    Opcode (line, optable), symbol_ (symbol)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;
  const EndDefineOpcode * is_enddefine_opcode () const;

protected:
  const std::string create_list_specific () const;
private:
  const Symbol * symbol_;
};

class LabelOpcode:public Opcode
{
public:
  LabelOpcode (const Label & label,
               int line,
               OpcodeTable * optable = 0):
    Opcode (line, optable), label_ (label)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;
  const LabelOpcode * is_label_opcode () const;

protected:
  const std::string create_list_specific () const;
private:
  const Label label_;
};

// Assembly language opcode and inlines.
class AssemblyOpcode:public Opcode
{
public:
  AssemblyOpcode (const std::string & assembly,
                  int line,
                  OpcodeTable * optable = 0):
    Opcode (line, optable), assembly_ (assembly)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;
  const AssemblyOpcode * is_assembly_opcode () const;

protected:
  const std::string create_list_specific () const;
private:
  const std::string assembly_;
};

// Register load and store.
class LoadValueOpcode:public Opcode
{
public:
  LoadValueOpcode (const Register & reg,
                   const Value value,
                   int line,
                   OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), value_ (value)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Value value_;
};

class LoadRegisterOpcode:public Opcode
{
public:
  LoadRegisterOpcode (const Register & reg,
                      const Register & with,
                      int line,
                      OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), with_ (with)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Register with_;
};

class LoadSymbolOpcode:public Opcode
{
public:
  LoadSymbolOpcode (const Register & reg,
                    const Symbol * symbol,
                    int line,
                    OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), symbol_ (symbol)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Symbol * symbol_;
};

class LoadSymbolIndirectOpcode:public Opcode
{
public:
  LoadSymbolIndirectOpcode (const Register & reg,
                            const Symbol * symbol,
                            int line,
                            OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), symbol_ (symbol)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Symbol * symbol_;
};

class LoadDataOpcode:public Opcode
{
public:
  LoadDataOpcode (const Register & reg,
                  const Data * data,
                  int line,
                  OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), data_ (data)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Data * data_;
};

class StoreSymbolIndirectOpcode:public Opcode
{
public:
  StoreSymbolIndirectOpcode (const Register & reg,
                             const Symbol * symbol,
                             int line,
                             OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), symbol_ (symbol)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Symbol * symbol_;
};

class PushOpcode:public Opcode
{
public:
  PushOpcode (const Register & reg,
              const Stack & stack,
              int line,
              OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), stack_ (stack)
  {
  }
  const Register & get_register () const;
  const Stack & get_stack () const;
  void generate (std::ostream & outs,
                 const Options & options) const;
  const PushOpcode * is_push_opcode () const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Stack stack_;
};

class PopOpcode:public Opcode
{
public:
  PopOpcode (const Register & reg,
             const Stack & stack,
             int line,
             OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), stack_ (stack)
  {
  }
  const Register & get_register () const;
  const Stack & get_stack () const;
  void generate (std::ostream & outs,
                 const Options & options) const;
  const PopOpcode * is_pop_opcode () const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Stack stack_;
};

// Inline push and pop getter functions.
inline const Register &
PushOpcode::get_register () const
{
  return reg_;
}

inline const Stack &
PushOpcode::get_stack () const
{
  return stack_;
}

inline const Register &
PopOpcode::get_register () const
{
  return reg_;
}

inline const Stack &
PopOpcode::get_stack () const
{
  return stack_;
}

// Add and subtract.
class AddValueOpcode:public Opcode
{
public:
  AddValueOpcode (const Register & reg,
                  const Value value,
                  int line,
                  OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), value_ (value)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Value value_;
};

class AddRegisterOpcode:public Opcode
{
public:
  AddRegisterOpcode (const Register & reg,
                     const Register & increment,
                     int line,
                     OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), increment_ (increment)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Register increment_;
};

class SubtractValueOpcode:public Opcode
{
public:
  SubtractValueOpcode (const Register & reg,
                       const Value value,
                       int line,
                       OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), value_ (value)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Value value_;
};

class SubtractRegisterOpcode:public Opcode
{
public:
  SubtractRegisterOpcode (const Register & reg,
                          const Register & decrement,
                          int line,
                          OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), decrement_ (decrement)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Register decrement_;
};

// Branches.
class JumpOpcode:public Opcode
{
public:
  JumpOpcode (const Label & label,
              int line,
              OpcodeTable * optable = 0):
    Opcode (line, optable), label_ (label)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;
  const JumpOpcode * is_jump_opcode () const;

protected:
  const std::string create_list_specific () const;
private:
  const Label label_;
};

class JumpZeroOpcode:public Opcode
{
public:
  JumpZeroOpcode (const Register & reg,
                  const Label & label,
                  int line,
                  OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), label_ (label)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Label label_;
};

class JumpNonZeroOpcode:public Opcode
{
public:
  JumpNonZeroOpcode (const Register & reg,
                     const Label & label,
                     int line,
                     OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), label_ (label)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Label label_;
};

class JumpEqualOpcode:public Opcode
{
public:
  JumpEqualOpcode (const Register & reg1,
                   const Register & reg2,
                   const Label & label,
                   int line,
                   OpcodeTable * optable = 0):
    Opcode (line, optable), reg1_ (reg1), reg2_ (reg2), label_ (label)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg1_;
  const Register reg2_;
  const Label label_;
};

class JumpNotEqualOpcode:public Opcode
{
public:
  JumpNotEqualOpcode (const Register & reg1,
                      const Register & reg2,
                      const Label & label,
                      int line,
                      OpcodeTable * optable = 0):
    Opcode (line, optable), reg1_ (reg1), reg2_ (reg2), label_ (label)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg1_;
  const Register reg2_;
  const Label label_;
};

class JumpGreaterEqualOpcode:public Opcode
{
public:
  JumpGreaterEqualOpcode (const Register & reg,
                          const Label & label,
                          int line,
                          OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), label_ (label)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Label label_;
};

class JumpLessEqualOpcode:public Opcode
{
public:
  JumpLessEqualOpcode (const Register & reg,
                       const Label & label,
                       int line,
                       OpcodeTable * optable = 0):
    Opcode (line, optable), reg_ (reg), label_ (label)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;

protected:
  const std::string create_list_specific () const;
private:
  const Register reg_;
  const Label label_;
};

// No-op, used for optimizing out opcodes.
class NoOpOpcode:public Opcode
{
public:
  NoOpOpcode (int line,
              OpcodeTable * optable = 0):
    Opcode (line, optable)
  {
  }
  void generate (std::ostream & outs,
                 const Options & options) const;
  const NoOpOpcode * is_noop_opcode () const;

protected:
  const std::string create_list_specific () const;
};

// Dynamic cast replacement functions.
inline const JumpOpcode *
Opcode::is_jump_opcode () const
{
  return 0;
}

inline const NoOpOpcode *
Opcode::is_noop_opcode () const
{
  return 0;
}

inline const LabelOpcode *
Opcode::is_label_opcode () const
{
  return 0;
}

inline const PushOpcode *
Opcode::is_push_opcode () const
{
  return 0;
}

inline const PopOpcode *
Opcode::is_pop_opcode () const
{
  return 0;
}

inline const DefineOpcode *
Opcode::is_define_opcode () const
{
  return 0;
}

inline const EndDefineOpcode *
Opcode::is_enddefine_opcode () const
{
  return 0;
}

inline const AssemblyOpcode *
Opcode::is_assembly_opcode () const
{
  return 0;
}

inline const AssemblyOpcode *
AssemblyOpcode::is_assembly_opcode () const
{
  return this;
}

inline const DefineOpcode *
DefineOpcode::is_define_opcode () const
{
  return this;
}

inline const EndDefineOpcode *
EndDefineOpcode::is_enddefine_opcode () const
{
  return this;
}

inline const LabelOpcode *
LabelOpcode::is_label_opcode () const
{
  return this;
}

inline const PushOpcode *
PushOpcode::is_push_opcode () const
{
  return this;
}

inline const PopOpcode *
PopOpcode::is_pop_opcode () const
{
  return this;
}

inline const JumpOpcode *
JumpOpcode::is_jump_opcode () const
{
  return this;
}

inline const NoOpOpcode *
NoOpOpcode::is_noop_opcode () const
{
  return this;
}

#endif
