// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <string>
#include <vector>
#include <algorithm>
#include <functional>

#include "register.h"
#include "stack.h"
#include "opcode.h"
#include "optable.h"

// Delete all opcode objects in the table, and destructor.
namespace {
  void
  dispose (const Opcode * opcode)
  {
    delete opcode;
  }
}
void
OpcodeTable::clear ()
{
  std::for_each (vector_.begin (), vector_.end (), dispose);
  vector_.erase (vector_.begin (), vector_.end ());
}
OpcodeTable::~OpcodeTable ()
{
  clear ();
}

// Add opcode objects sequentially to the vector.
void
OpcodeTable::add (Opcode * opcode)
{
  vector_.push_back (opcode);
}

// Pretty-print to create the opcode table listing.
namespace {
  void
  print_out (const Opcode * opcode,
             std::ostream * outs)
  {
    *outs << opcode->create_list_entry ();
  }
}
void
OpcodeTable::create_listing (std::ostream & outs) const
{
  outs << "Intermediate code:" << std::endl;
  if (vector_.empty ())
    {
      outs << " (empty)" << std::endl;
      return;
    }

  outs << Opcode::create_list_header ();
  std::for_each (vector_.begin (), vector_.end (),
                 std::bind2nd (std::ptr_fun (print_out), &outs));
}

// Optimization functions, replace adjacent push-pop pairs with a possible
// shorter representation.
bool
OpcodeTable::optimize_pair (OpcodeTableVectorIterator_mutable push_iter,
                            OpcodeTableVectorIterator_mutable pop_iter)
{
  const PushOpcode * push = (*push_iter)->is_push_opcode ();
  const PopOpcode * pop = (*pop_iter)->is_pop_opcode ();

  // Replace if the pop pops the stack just pushed.
  if (push->get_stack ().equals (pop->get_stack ()))
    {
      // Either eliminate, or transform into register to register move.
      Opcode * replacement;
      if (push->get_register ().equals (pop->get_register ()))
        replacement = new NoOpOpcode (push->get_line ());
      else
        replacement = new LoadRegisterOpcode (pop->get_register (),
                                              push->get_register (),
                                              push->get_line ());
      replacement->set_optimizes (*push_iter);
      *push_iter = replacement;

      // Always eliminate the second opcode of the pair.
      replacement = new NoOpOpcode (pop->get_line ());
      replacement->set_optimizes (*pop_iter);
      *pop_iter = replacement;

      return true;
    }

  return false;
}

bool
OpcodeTable::optimize_pass ()
{
  bool status = false;

  OpcodeTableVectorIterator_mutable iter, pop_iter;
  for (iter = vector_.begin (); iter < vector_.end (); iter = pop_iter)
    {
      // Find the next push opcode.
      OpcodeTableVectorIterator_mutable push_iter;
      push_iter = std::find_if (iter, vector_.end (),
                                std::mem_fun (&Opcode::is_push_opcode));
      if (push_iter == vector_.end ())
        break;

      // If followed by a pop, try to optimize the pair.
      pop_iter = push_iter + 1;
      if (pop_iter == vector_.end ())
        break;

      if ((*pop_iter)->is_pop_opcode ())
          status |= optimize_pair (push_iter, pop_iter);
    }

  return status;
}

// Currently simply a single optimization pass over opcodes.
void
OpcodeTable::optimize_code ()
{
  optimize_pass ();
}

// Unreachability check, finds occurrences of an unconditional jump that is
// followed by a non-no-op, non-label opcode.
void
OpcodeTable::unreachable_check (std::string source_path) const
{
  OpcodeTableVectorIterator iter, jump_iter;
  for (iter = vector_.begin (); iter < vector_.end (); iter = jump_iter + 1)
    {
      // Find the next unconditional jump.
      jump_iter = std::find_if (iter, vector_.end (),
                                std::mem_fun (&Opcode::is_jump_opcode));
      if (jump_iter == vector_.end ())
        break;

      // Find the non-no-op after the jump.
      OpcodeTableVectorIterator next_iter;
      next_iter = std::find_if (jump_iter + 1, vector_.end (),
                                std::not1
                                (std::mem_fun (&Opcode::is_noop_opcode)));
      if (next_iter == vector_.end ())
        break;

      // If not a label, this code is unreachable.
      if (!(*next_iter)->is_label_opcode ())
        {
          std::cout << source_path << ':' << (*next_iter)->get_line ()
            << ": warning: unreachable code, skipped by line "
            << (*jump_iter)->get_line () << std::endl;
        }
    }
}
