// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <cstdio>
#include <cstdlib>
#include <climits>
#include <cerrno>
#include <cassert>
#include <string>
#include <iostream>
#include <sstream>
#include <stack>

#include "data.h"
#include "dattable.h"
#include "symbol.h"
#include "symtable.h"
#include "opcode.h"
#include "optable.h"
#include "operand.h"
#include "parser.h"

// Global declaration of pointer to active parser, linked to by yacc/lex
// modules, and external declaration of the bits of yacc/lex we call/use.
Parser * _parser = 0;
extern void yyrestart (std::FILE *);
extern int yyparse ();
extern const char * yylval;

// Static count of max tolerable errors before giving up.
const int Parser::MAX_ALLOWED_ERRORS = 20;

// Front-end parser function, driver for yacc/lex code.
bool
Parser::parse (std::string source_path,
               std::FILE * stream,
               DataTable * dattable,
               SymbolTable * symtable,
               OpcodeTable * optable)
{
  source_path_ = source_path;

  dattable_ = dattable;
  symtable_ = symtable;
  optable_ = optable;

  parse_setup ();

  _parser = this;
  yyrestart (stream);
  int parse_status = yyparse ();
  _parser = 0;

  parse_cleanup ();

  dattable_ = 0;
  symtable_ = 0;
  optable_ = 0;

  return (parse_status == 0) && (error_count_ == 0);
}

void
Parser::parse_setup ()
{
  line_number_ = 1;
  error_count_ = 0;
  current_definition_ = 0;

  label_sequence_ = 1;
  exit_label_ = 0;

  implicit_main_ = false;
}

void
Parser::parse_cleanup ()
{
  if (symtable_->lookup ("main"))
    {
      add_main_symbol ("_argc");
      add_main_symbol ("_argv");
      add_main_symbol ("_envp");
    }

  if (error_count_ == 0)
    {
      if (!(label_stack_.empty () && break_stack_.empty ()))
        {
          parse_error ()
            << "internal error: parse stacks not emptied" << std::endl;
        }

      if (current_definition_)
        {
          parse_error ()
            << "internal error: definition not exited" << std::endl;
        }
      if (exit_label_)
        {
          parse_error ()
            << "internal error: definition end missed" << std::endl;
        }

    if (dattable_->is_empty ()
        && symtable_->is_empty ()
        && optable_->is_empty ())
      {
        parse_error ()
          << "warning: file contained no symbols or code" << std::endl;
      }
    }
}

void
Parser::add_main_symbol (const char * s)
{
  if (const Symbol * symbol = symtable_->lookup (s))
    {
      if (!symbol->is_external_symbol ())
        {
          parse_error () << "'" << s << "' defined in main module"
            << " (line " << symbol->get_line () << ")" << std::endl;
        }
    }
  else
    new VariableSymbol (s, line_number_, *symtable_);
}

// Parser error reporting, called from yacc/lex or internally only.
std::ostream &
Parser::parse_error ()
{
  ++error_count_;
  std::cerr << source_path_ << ':';
  if (_parser && yylval)
    std::cerr << line_number_ << ':' << " near '" << yylval << "', ";
  else
    std::cerr << ' ';
  return std::cerr;
}

std::ostream &
Parser::parse_warning ()
{
  std::cerr << source_path_ << ':';
  if (_parser && yylval)
    std::cerr << line_number_ << ':' << " near '" << yylval << "', ";
  else
    std::cerr << ' ';
  return std::cerr;
}

// Detect implicit main function.
void
Parser::detect_main ()
{
  if (current_definition_ || implicit_main_)
    return;

  if (const Symbol * symbol = symtable_->lookup ("main"))
    {
      parse_error () << "'main' both defined and implicit"
        << " (line " << symbol->get_line () << ")" << std::endl;
    }
  else
    new WordSymbol ("main", line_number_, *symtable_);

  implicit_main_ = true;
}

// Parse terminal handlers, all called from yacc/lex only.
void
Parser::f_definedvariable (const char * s)
{
  if (const Symbol * symbol = symtable_->lookup (s))
    {
      if (symbol->is_external_symbol ())
        {
          parse_error () << "symbol '" << s << "' already implicitly external"
            << " (line " << symbol->get_line () << ")" << std::endl;
        }
      else
        {
          parse_error () << "multiple symbol definition '" << s
            << "' (line " << symbol->get_line () << ")" << std::endl;
        }
      return;
    }

  new VariableSymbol (s, line_number_, *symtable_);
}

void
Parser::f_definedarray (const char * s, const char * ssize, const char * type)
{
  std::istringstream ins (s);

  std::string ss, stype;
  ins >> ss;

  if (const Symbol * symbol = symtable_->lookup (ss))
    {
      if (symbol->is_external_symbol ())
        {
          parse_error () << "symbol '" << ss << "' already implicitly external"
            << " (line " << symbol->get_line () << ")" << std::endl;
        }
      else
        {
          parse_error () << "multiple symbol definition '" << ss
            << "' (line " << symbol->get_line () << ")" << std::endl;
        }
      return;
    }

  errno = 0;
  int size = std::strtol (ssize, 0, 0);
  if (errno == ERANGE)
    {
      parse_error () << "array size '" << ssize
        << "' out of range" << std::endl;
      return;
    }
  if (size <= 0)
    {
      parse_error () << "array '" << ss
        << "' cannot have size zero or less" << std::endl;
      return;
    }

  std::istringstream ins2 (type);
  ins2 >> stype;
  if (stype != "CHARS" && stype != "CELLS")
    {
      parse_error () << "parse error, "
        << "expecting 'CHARS' or 'CELLS'" << std::endl;
      return;
    }

  VariableSymbol::Units units = (stype == "CHARS")
                  ? VariableSymbol::CHARACTER : VariableSymbol::CELL;
  new VariableSymbol (ss, size, units, line_number_, *symtable_);
}

void
Parser::f_definedarray (const char * s, const char * ssize)
{
  f_definedarray (s, ssize, "CHARS");
}

void
Parser::f_definedconstant (const char * s)
{
  if (const Symbol * symbol = symtable_->lookup (s))
    {
      if (symbol->is_external_symbol ())
        {
          parse_error () << "symbol '" << s << "' already implicitly external"
            << " (line " << symbol->get_line () << ")" << std::endl;
        }
      else
        {
          parse_error () << "multiple symbol definition '" << s
            << "' (line " << symbol->get_line () << ")" << std::endl;
        }
      return;
    }

  const Symbol * symbol = new ConstantSymbol (s, line_number_, *symtable_);
  detect_main ();
  new PopOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
  new StoreSymbolIndirectOpcode (Register::R0, symbol, line_number_, optable_);
}

void
Parser::f_definedword (const char * s)
{
  if (const Symbol * symbol = symtable_->lookup (s))
    {
      if (symbol->is_external_symbol ())
        {
          parse_error () << "symbol '" << s << "' already implicitly external"
            << " (line " << symbol->get_line () << ")" << std::endl;
        }
      else
        {
          parse_error () << "multiple symbol definition '" << s
            << "' (line " << symbol->get_line () << ")" << std::endl;
        }

      assert (exit_label_ == 0);
      exit_label_ = label_sequence_;

      label_stack_.push (Label (label_sequence_++));
      return;
    }

  assert (exit_label_ == 0);
  exit_label_ = label_sequence_;

  label_stack_.push (Label (label_sequence_++));

  DefinableSymbol * symbol = new WordSymbol (s, line_number_, *symtable_);
  current_definition_ = symbol;
  new DefineOpcode (symbol, line_number_, optable_);
}

void
Parser::f_semicolon ()
{
  assert (!label_stack_.empty ());
  new LabelOpcode (label_stack_.top (), line_number_, optable_);
  label_stack_.pop ();

  assert (exit_label_ != 0);
  exit_label_ = 0;

  if (current_definition_)
    {
      new EndDefineOpcode (current_definition_, line_number_, optable_);
      current_definition_ = 0;
    }
}

void
Parser::f_exit ()
{
  detect_main ();
  new JumpOpcode (Label (exit_label_), line_number_, optable_);
}

void
Parser::f_pstring (const char * s)
{
  const CallableSymbol * csymbol;
  if (const Symbol * symbol = symtable_->lookup ("TYPE"))
    {
      csymbol = symbol->is_callable_symbol ();
      if (csymbol == 0)
        {
          parse_error () << "'TYPE' is defined as a variable"
            << "' (line " << symbol->get_line () << ")" << std::endl;
        }
    }
  else
    csymbol = new ExternalSymbol ("TYPE", line_number_, *symtable_);

  std::string ss = s;
  ss.erase (ss.begin (), ss.begin () + 3);
  ss.erase (ss.end () - 1);
  const Data * data = new StringData (ss.size (), ss.c_str (),
                                      line_number_, *dattable_);

  detect_main ();
  new LoadDataOpcode (Register::R0, data, line_number_, optable_);
  new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);

  new LoadValueOpcode (Register::R0, Value (ss.size ()),
                       line_number_, optable_);
  new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);

  new CallOpcode (csymbol, line_number_, optable_);
}

void
Parser::f_cstring (const char * s)
{
  std::string ss = s;
  ss.erase (ss.begin (), ss.begin () + 3);
  ss.erase (ss.end () - 1);

  if (ss.size () > UCHAR_MAX)
    {
      parse_error () << "string is longer than " << UCHAR_MAX
        << " characters" << std::endl;
      return;
    }

  const Data * data = new CountedStringData (ss.size (), ss.c_str (),
                                             line_number_, *dattable_);

  detect_main ();
  new LoadDataOpcode (Register::R0, data, line_number_, optable_);
  new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
}

void
Parser::f_word (const char * s)
{
  const Symbol * symbol = symtable_->lookup (s);
  if (!symbol)
    symbol = new ExternalSymbol (s, line_number_, *symtable_);

  if (symbol->is_variable_symbol ())
    {
      detect_main ();
      new LoadSymbolOpcode (Register::R0, symbol, line_number_, optable_);
      new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
      return;
    }

  else if (symbol->is_constant_symbol ())
    {
      detect_main ();
      new LoadSymbolIndirectOpcode (Register::R0, symbol,
                                    line_number_, optable_);
      new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
      return;
    }

  const CallableSymbol * csymbol = symbol->is_callable_symbol ();
  if (csymbol)
    {
      detect_main ();
      new CallOpcode (csymbol, line_number_, optable_);
      return;
    }

  parse_error () << "internal error: invalid symbol type" << std::endl;
}

void
Parser::f_object_address (const char * s)
{
  std::istringstream ins (s);

  std::string junk, ss;
  ins >> junk >> ss;

  const Symbol * symbol = symtable_->lookup (ss);
  if (!symbol)
    symbol = new ExternalSymbol (ss, line_number_, *symtable_);

  detect_main ();
  new LoadSymbolOpcode (Register::R0, symbol, line_number_, optable_);
  new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
}

void
Parser::f_char_literal (const char * s)
{
  std::istringstream ins (s);

  std::string junk, literal;
  ins >> junk >> literal;

  if (literal.size () > 1)
    {
      parse_warning () << "warning: char literal '"
        << literal << "' >1 character" << std::endl;
    }

  detect_main ();
  new LoadValueOpcode (Register::R0, Value (literal[0]),
                       line_number_, optable_);
  new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
}

void
Parser::f_integer_literal (const char * s)
{
  union {
    int i;
    unsigned int u;
  } literal;

  errno = 0;
  literal.i = std::strtol (s, 0, 0);
  if (errno == ERANGE)
    {
      if (s[0] == '-')
        {
          parse_error () << "integer literal '" << s
            << "' out of range" << std::endl;
          return;
        }

      errno = 0;
      literal.u = std::strtoul (s, 0, 0);
      if (errno == ERANGE)
        {
          parse_error () << "integer literal '" << s
            << "' out of range" << std::endl;
          return;
        }
    }

  detect_main ();
  new LoadValueOpcode (Register::R0, Value (literal.i),
                       line_number_, optable_);
  new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
}

void
Parser::f_float_literal (const char * s)
{
  if (sizeof (float) != sizeof (int))
    parse_error () << "internal error: float and int sizes differ" << std::endl;

  union {
    int i;
    float f;
  } literal;

  errno = 0;
  literal.f = std::strtof (s, 0);
  if (errno == ERANGE)
    {
      parse_error () << "float literal '" << s
        << "' out of range" << std::endl;
      return;
    }

  detect_main ();
  new LoadValueOpcode (Register::R0, Value (literal.i),
                       line_number_, optable_);
  new PushOpcode (Register::R0, Stack::FloatStack, line_number_, optable_);
}

void
Parser::f_string_literal (const char * s)
{
  std::string ss = s;
  ss.erase (ss.begin (), ss.begin () + 3);
  ss.erase (ss.end () - 1);
  const Data * data = new StringData (ss.size (), ss.c_str (),
                                      line_number_, *dattable_);

  detect_main ();
  new LoadDataOpcode (Register::R0, data, line_number_, optable_);
  new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);

  new LoadValueOpcode (Register::R0, Value (ss.size ()),
                       line_number_, optable_);
  new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
}

void
Parser::f_if ()
{
  label_stack_.push (Label (label_sequence_++));

  detect_main ();
  new PopOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
  new JumpZeroOpcode (Register::R0, label_stack_.top (),
                      line_number_, optable_);
}

void
Parser::f_then ()
{
  detect_main ();
  assert (!label_stack_.empty ());
  new LabelOpcode (label_stack_.top (), line_number_, optable_);
  label_stack_.pop ();
}

void
Parser::f_else ()
{
  assert (!label_stack_.empty ());
  Label over = label_stack_.top ();
  label_stack_.pop ();

  label_stack_.push (Label (label_sequence_++));

  detect_main ();
  new JumpOpcode (label_stack_.top (), line_number_, optable_);
  new LabelOpcode (over, line_number_, optable_);
}

void
Parser::f_do ()
{
  detect_main ();
  new PopOpcode (Register::R1, Stack::DataStack, line_number_, optable_);
  new PopOpcode (Register::R0, Stack::DataStack, line_number_, optable_);

  label_stack_.push (Label (label_sequence_++));

  new PushOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);

  new SubtractRegisterOpcode (Register::R1, Register::R0,
                              line_number_, optable_);
  new PushOpcode (Register::R1, Stack::ReturnStack, line_number_, optable_);

  label_stack_.push (Label (label_sequence_++));

  new LabelOpcode (label_stack_.top (), line_number_, optable_);

  label_stack_.push (Label (label_sequence_++));
  break_stack_.push (label_stack_.top ());
}

void
Parser::f_query_do ()
{
  detect_main ();
  new PopOpcode (Register::R1, Stack::DataStack, line_number_, optable_);
  new PopOpcode (Register::R0, Stack::DataStack, line_number_, optable_);

  label_stack_.push (Label (label_sequence_++));

  new JumpEqualOpcode (Register::R0, Register::R1, label_stack_.top (),
                       line_number_, optable_);

  new PushOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);

  new SubtractRegisterOpcode (Register::R1, Register::R0,
                              line_number_, optable_);
  new PushOpcode (Register::R1, Stack::ReturnStack, line_number_, optable_);

  label_stack_.push (Label (label_sequence_++));

  new LabelOpcode (label_stack_.top (), line_number_, optable_);

  label_stack_.push (Label (label_sequence_++));
  break_stack_.push (label_stack_.top ());
}

void
Parser::f_loop ()
{
  assert (!label_stack_.empty ());
  Label end_label = label_stack_.top ();
  label_stack_.pop ();

  detect_main ();
  new PopOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);
  new AddValueOpcode (Register::R0, Value (1), line_number_, optable_);
  new PushOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);

  assert (!label_stack_.empty ());
  new JumpNonZeroOpcode (Register::R0, label_stack_.top (),
                         line_number_, optable_);
  label_stack_.pop ();

  new LabelOpcode (end_label, line_number_, optable_);
  new PopOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);
  new PopOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);

  assert (!label_stack_.empty ());
  new LabelOpcode (label_stack_.top (), line_number_, optable_);
  label_stack_.pop ();

  assert (!break_stack_.empty ());
  break_stack_.pop ();
}

void
Parser::f_plus_loop ()
{
  assert (!label_stack_.empty ());
  Label end_label = label_stack_.top ();
  label_stack_.pop ();

  detect_main ();
  new PopOpcode (Register::R1, Stack::DataStack, line_number_, optable_);
  new PopOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);

  new AddRegisterOpcode (Register::R0, Register::R1, line_number_, optable_);
  new PushOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);

  assert (!label_stack_.empty ());
  new JumpGreaterEqualOpcode (Register::R0, label_stack_.top (),
                              line_number_, optable_);
  label_stack_.pop ();

  new LabelOpcode (end_label, line_number_, optable_);
  new PopOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);
  new PopOpcode (Register::R0, Stack::ReturnStack, line_number_, optable_);

  assert (!label_stack_.empty ());
  new LabelOpcode (label_stack_.top (), line_number_, optable_);
  label_stack_.pop ();

  assert (!break_stack_.empty ());
  break_stack_.pop ();
}

void
Parser::f_leave ()
{
  if (break_stack_.empty ())
    {
      parse_error () << "'LEAVE' invalid outside a loop" << std::endl;
      return;
    }

  detect_main ();
  new JumpOpcode (break_stack_.top (), line_number_, optable_);
}

void
Parser::f_begin ()
{
  detect_main ();
  label_stack_.push (Label (label_sequence_++));
  new LabelOpcode (label_stack_.top (), line_number_, optable_);

  label_stack_.push (Label (label_sequence_++));
  break_stack_.push (label_stack_.top ());
}

void
Parser::f_until ()
{
  assert (!label_stack_.empty ());
  Label end_label = label_stack_.top ();
  label_stack_.pop ();

  detect_main ();
  new PopOpcode (Register::R0, Stack::DataStack, line_number_, optable_);

  assert (!label_stack_.empty ());
  new JumpZeroOpcode (Register::R0, label_stack_.top (),
                      line_number_, optable_);
  label_stack_.pop ();

  new LabelOpcode (end_label, line_number_, optable_);

  assert (!break_stack_.empty ());
  break_stack_.pop ();
}

void
Parser::f_again ()
{
  assert (!label_stack_.empty ());
  Label end_label = label_stack_.top ();
  label_stack_.pop ();

  detect_main ();
  assert (!label_stack_.empty ());
  new JumpOpcode (label_stack_.top (), line_number_, optable_);
  label_stack_.pop ();

  new LabelOpcode (end_label, line_number_, optable_);

  assert (!break_stack_.empty ());
  break_stack_.pop ();
}

void
Parser::f_while ()
{
  if (break_stack_.empty ())
    {
      parse_error () << "'WHILE' invalid outside a loop" << std::endl;
      return;
    }

  detect_main ();
  new PopOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
  new JumpZeroOpcode (Register::R0, break_stack_.top (),
                      line_number_, optable_);
}

void
Parser::f_repeat ()
{
  assert (!label_stack_.empty ());
  Label end_label = label_stack_.top ();
  label_stack_.pop ();

  detect_main ();
  assert (!label_stack_.empty ());
  new JumpOpcode (label_stack_.top (), line_number_, optable_);
  label_stack_.pop ();

  new LabelOpcode (end_label, line_number_, optable_);

  assert (!break_stack_.empty ());
  break_stack_.pop ();
}

void
Parser::f_case ()
{
  detect_main ();
  label_stack_.push (Label (label_sequence_++));
  break_stack_.push (label_stack_.top ());
}

void
Parser::f_of ()
{
  detect_main ();
  new PopOpcode (Register::R1, Stack::DataStack, line_number_, optable_);
  new PopOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
  new PushOpcode (Register::R0, Stack::DataStack, line_number_, optable_);

  label_stack_.push (Label (label_sequence_++));

  new JumpNotEqualOpcode (Register::R0, Register::R1,
                          label_stack_.top (), line_number_, optable_);

  new PopOpcode (Register::R0, Stack::DataStack, line_number_, optable_);
}

void
Parser::f_endof ()
{
  detect_main ();
  assert (!break_stack_.empty ());
  new JumpOpcode (break_stack_.top (), line_number_, optable_);

  assert (!label_stack_.empty ());
  new LabelOpcode (label_stack_.top (), line_number_, optable_);
  label_stack_.pop ();
}

void
Parser::f_endcase ()
{
  detect_main ();
  new PopOpcode (Register::R0, Stack::DataStack, line_number_, optable_);

  assert (!label_stack_.empty ());
  new LabelOpcode (label_stack_.top (), line_number_, optable_);
  label_stack_.pop ();

  assert (!break_stack_.empty ());
  break_stack_.pop ();
}

void
Parser::f_recurse ()
{
  if (!current_definition_)
    {
      parse_error () << "'RECURSE' invalid outside a definition" << std::endl;
      return;
    }

  detect_main ();
  new CallOpcode (current_definition_, line_number_, optable_);
}

void
Parser::f_codeword (const char * s)
{
  if (const Symbol * symbol = symtable_->lookup (s))
    {
      if (symbol->is_external_symbol ())
        {
          parse_error () << "symbol '" << s << "' already implicitly external"
            << " (line " << symbol->get_line () << ")" << std::endl;
        }
      else
        {
          parse_error () << "multiple symbol definition '" << s
            << "' (line " << symbol->get_line () << ")" << std::endl;
        }
      return;
    }

  DefinableSymbol * symbol = new CodewordSymbol (s, line_number_, *symtable_);
  current_definition_ = symbol;
  new DefineOpcode (symbol, line_number_, optable_);
}

void
Parser::f_code (const char * s)
{
  int i;
  for (i = 0;
      s[i] && s[i] != ';' && s[i] != '#'; ++i)
    {
      if (s[i] == '"')
        {
          for (++i; s[i] && s[i] != '"'; ++i)
            {
              if (s[i] == '\\')
                ++i;
            }
        }
    }
  for (--i; i >= 0 && isspace (s[i]); --i)
    ;

  if (i > 0)
    {
      std::string ss = std::string (s, 0, i + 1);
      new AssemblyOpcode (ss, line_number_, optable_);
    }
}

void
Parser::f_end_code ()
{
  if (current_definition_)
    {
      new EndDefineOpcode (current_definition_, line_number_, optable_);
      current_definition_ = 0;
    }
}
