// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#ifndef _VNPFORTH_PARSER_H
#define _VNPFORTH_PARSER_H

#include <cstdio>
#include <iostream>
#include <string>
#include <stack>

#include "operand.h"

class DataTable;
class SymbolTable;
class OpcodeTable;
class CommandLine;
class DefinableSymbol;

// Parser, builds a program from source.
class Parser
{
public:
  static const int MAX_ALLOWED_ERRORS;

  Parser ();
  bool parse (std::string source_path,
              std::FILE * stream,
              DataTable * dattable,
              SymbolTable * symtable,
              OpcodeTable * optable);

  // Parser terminals, public to be callable by non-OO yacc/lex modules.
  void next_line ();
  std::ostream & parse_error ();
  bool excessive_errors () const;
  void f_definedvariable (const char * s);
  void f_definedarray (const char * s, const char * ssize, const char * type);
  void f_definedarray (const char * s, const char * ssize);
  void f_definedconstant (const char * s);
  void f_definedword (const char * s);
  void f_semicolon ();
  void f_exit ();
  void f_pstring (const char * s);
  void f_cstring (const char * s);
  void f_word (const char * s);
  void f_object_address (const char * s);
  void f_char_literal (const char * s);
  void f_integer_literal (const char * s);
  void f_float_literal (const char * s);
  void f_string_literal (const char * s);
  void f_if ();
  void f_then ();
  void f_else ();
  void f_do ();
  void f_query_do ();
  void f_loop ();
  void f_plus_loop ();
  void f_leave ();
  void f_begin ();
  void f_until ();
  void f_again ();
  void f_while ();
  void f_repeat ();
  void f_case ();
  void f_of ();
  void f_endof ();
  void f_endcase ();
  void f_recurse ();
  void f_codeword (const char * s);
  void f_code (const char * s);
  void f_end_code ();

private:
  void parse_setup ();
  void parse_cleanup ();
  void add_main_symbol (const char * s);

  std::ostream & parse_warning ();

  std::string source_path_;
  DataTable * dattable_;
  SymbolTable * symtable_;
  OpcodeTable * optable_;

  int line_number_;
  int error_count_;

  const DefinableSymbol * current_definition_;

  int label_sequence_;
  int exit_label_;
  std::stack<Label> label_stack_;
  std::stack<Label> break_stack_;

  bool implicit_main_;
  void detect_main ();
};

// Inline function definitions.
inline
Parser::Parser ():
  dattable_ (0), symtable_ (0), optable_ (0), line_number_ (0),
  error_count_ (0), current_definition_ (0), label_sequence_ (0),
  exit_label_ (0), implicit_main_ (false)
{
}

inline void
Parser::next_line ()
{
  ++line_number_;
}

inline bool
Parser::excessive_errors () const
{
  return error_count_ >= MAX_ALLOWED_ERRORS;
}

#endif
