// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <ctime>
#include <cstdio>
#include <sys/utsname.h>
#include <iostream>
#include <string>

#include "dattable.h"
#include "symtable.h"
#include "optable.h"
#include "parser.h"
#include "srcfile.h"
#include "cmdline.h"
#include "options.h"
#include "program.h"

// Parse a source file into the program.
bool
Program::build (const std::string & source_path,
                const CommandLine & commandline)
{
  SourceFile source = SourceFile (source_path);
  try {
    source.open ();
  }
  catch (std::string & error) {
    std::cerr << commandline.get_program_name ()
      << ": " << source_path << ": " << error << std::endl;
    return false;
  }

  source_path_ = source_path;
  options_ = commandline.get_options ();

  datatable_.clear ();
  symtable_.clear ();
  optable_.clear ();

  Parser parser;
  int status = parser.parse (source_path, source.get_stream (),
                             &datatable_, &symtable_, &optable_);

  if (status && options_.optimize_code ())
    {
      optable_.optimize_code ();
    }

  optable_.unreachable_check (source_path);

  return status;
}

// Front end to assembly code generation.
void
Program::generate (std::ostream & outs) const
{
  generate_preamble (outs);
  datatable_.generate (outs);
  symtable_.generate (outs, options_);
  optable_.generate (outs, options_, symtable_);
  generate_postamble (outs);
}

// Pretty-print function to create intermediate store listing.
void
Program::create_listing (std::ostream & outs) const
{
  outs << "Forth 1.4 [" << __DATE__ << ' ' << __TIME__ << ", ";
  utsname ubuf;
  outs << (uname (&ubuf) == 0 ? ubuf.sysname : "Unknown");
  outs << "] compilation listing" << std::endl << std::endl;

  outs << "Source file:   " << source_path_ << std::endl;
  std::time_t now;
  std::time (&now);
  outs << "Compiled on:   " << asctime (localtime (&now));
  outs << "Compile flags:"
    << (options_.include_debugging () ? " Debug" : "")
    << (options_.include_profiling () ? " Profile" : "")
    << (options_.position_independent () ? " PIC" : "")
    << (options_.optimize_code () ? " Optimize" : "")
    << (options_.save_intermediate () ? " Intermediate" : "")
    << (options_.save_assembly () ? " Assembly" : "")
    << std::endl << std::endl;

  datatable_.create_listing (outs);
  outs << std::endl;

  symtable_.create_listing (outs);
  outs << std::endl;

  optable_.create_listing (outs);
  outs << std::endl;
}
