// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#ifndef _VNPFORTH_SYMBOL_H
#define _VNPFORTH_SYMBOL_H

#include <cstddef>
#include <string>

class SymbolTable;
class Options;
class VariableSymbol;
class ConstantSymbol;
class CallableSymbol;
class ExternalSymbol;
class DefinableSymbol;
class WordSymbol;
class CodewordSymbol;

// Abstract symbol class, assigns a unique sequential id on construction.  
class Symbol
{
public:
  virtual ~Symbol ();

  const std::string get_name () const;
  int get_id () const;
  int get_line () const;

  static const std::string create_list_header ();
  virtual const std::string create_list_entry () const;

  virtual void generate (std::ostream & outs,
                         const Options & options) const;

  virtual const VariableSymbol * is_variable_symbol () const;
  virtual const ConstantSymbol * is_constant_symbol () const;
  virtual const CallableSymbol * is_callable_symbol () const;
  virtual const ExternalSymbol * is_external_symbol () const;
  virtual const DefinableSymbol * is_definable_symbol () const;
  virtual const WordSymbol * is_word_symbol () const;
  virtual const CodewordSymbol * is_codeword_symbol () const;

protected:
  Symbol (const std::string & name,
          int line,
          const std::string type,
          SymbolTable & symtable);

  const std::string get_type () const;

private:
  static int count;
  static int instances;

  const std::string name_;
  const int id_;
  const int line_;
  const std::string type_;

};

// Inline function definitions.
inline const std::string
Symbol::get_name () const
{
  return name_;
}

inline int
Symbol::get_id () const
{
  return id_;
}

inline int
Symbol::get_line () const
{
  return line_;
}

inline const std::string
Symbol::get_type () const
{
  return type_;
}

// Concrete symbol classes.
class VariableSymbol:public Symbol
{
public:
  enum Units { CHARACTER, CELL };

  VariableSymbol (const std::string & name,
                  size_t size,
                  Units units,
                  int line,
                  SymbolTable & symtable):
    Symbol (name, line, "Variable", symtable), size_ (size), units_ (units)
  {
  }
  VariableSymbol (const std::string & name,
                  int line,
                  SymbolTable & symtable):
    Symbol (name, line, "Variable", symtable), size_ (1), units_ (CELL)
  {
  }

  const std::string create_list_entry () const;

  void generate (std::ostream & outs,
                 const Options & options) const;
  const VariableSymbol * is_variable_symbol () const;

private:
  const size_t size_;
  const Units units_;
};

class ConstantSymbol:public Symbol
{
public:
  ConstantSymbol (const std::string & name,
                  int line,
                  SymbolTable & symtable):
    Symbol (name, line, "Constant", symtable)
  {
  }

  void generate (std::ostream & outs,
                 const Options & options) const;
  const ConstantSymbol * is_constant_symbol () const;

};

// Abstract symbol for objects suitable for function calls.
class CallableSymbol:public Symbol
{
protected:
  CallableSymbol (const std::string & name,
                  int line,
                  const std::string type,
                  SymbolTable & symtable):
    Symbol (name, line, type, symtable)
  {
  }
  const CallableSymbol * is_callable_symbol () const;
};

// Concrete callable symbol classes.
class ExternalSymbol:public CallableSymbol
{
public:
  ExternalSymbol (const std::string & name,
                  int line,
                  SymbolTable & symtable):
    CallableSymbol (name, line, "External", symtable)
  {
  }

  void generate (std::ostream & outs,
                 const Options & options) const;
  const ExternalSymbol * is_external_symbol () const;
};

class DefinableSymbol:public CallableSymbol
{
protected:
  DefinableSymbol (const std::string & name,
                   int line,
                   const std::string type,
                   SymbolTable & symtable):
    CallableSymbol (name, line, type, symtable)
  {
  }
  const DefinableSymbol * is_definable_symbol () const;
};

class WordSymbol:public DefinableSymbol
{
public:
  WordSymbol (const std::string & name,
              int line,
              SymbolTable & symtable):
    DefinableSymbol (name, line, "Word", symtable)
  {
  }
  const WordSymbol * is_word_symbol () const;
};

class CodewordSymbol:public DefinableSymbol
{
public:
  CodewordSymbol (const std::string & name,
                  int line,
                  SymbolTable & symtable):
    DefinableSymbol (name, line, "Codeword", symtable)
  {
  }
  const CodewordSymbol * is_codeword_symbol () const;
};

// Dynamic cast replacement functions.
inline const VariableSymbol *
Symbol::is_variable_symbol () const
{
  return 0;
}

inline const ConstantSymbol *
Symbol::is_constant_symbol () const
{
  return 0;
}

inline const CallableSymbol *
Symbol::is_callable_symbol () const
{
  return 0;
}

inline const ExternalSymbol *
Symbol::is_external_symbol () const
{
  return 0;
}

inline const DefinableSymbol *
Symbol::is_definable_symbol () const
{
  return 0;
}

inline const WordSymbol *
Symbol::is_word_symbol () const
{
  return 0;
}

inline const CodewordSymbol *
Symbol::is_codeword_symbol () const
{
  return 0;
}

inline const VariableSymbol *
VariableSymbol::is_variable_symbol () const
{
  return this;
}

inline const ConstantSymbol *
ConstantSymbol::is_constant_symbol () const
{
  return this;
}

inline const CallableSymbol *
CallableSymbol::is_callable_symbol () const
{
  return this;
}

inline const ExternalSymbol *
ExternalSymbol::is_external_symbol () const
{
  return this;
}

inline const DefinableSymbol *
DefinableSymbol::is_definable_symbol () const
{
  return this;
}

inline const WordSymbol *
WordSymbol::is_word_symbol () const
{
  return this;
}

inline const CodewordSymbol *
CodewordSymbol::is_codeword_symbol () const
{
  return this;
}

#endif
