// vi: set ts=2 shiftwidth=2 expandtab:
//
// VNPForth - Compiled native Forth for IA32 Linux systems
// Copyright (C) 2005  Simon Baldwin (simon_baldwin@yahoo.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <string>
#include <algorithm>
#include <functional>

#include "symbol.h"
#include "mangler.h"
#include "symtable.h"

// Delete all symbol objects in the table, and destructor.
namespace {
  void
  dispose (SymbolTable::SymbolTableMapPair entry)
  {
    delete entry.second;
  }
}
void
SymbolTable::clear ()
{
  std::for_each (map_.begin (), map_.end (), dispose);
  map_.erase (map_.begin (), map_.end ());
}
SymbolTable::~SymbolTable ()
{
  clear ();
}

// Add and lookup symbols; name lookup mangles before searching, as symbol
// names are always mangled on symbol construction.
void
SymbolTable::add (const Symbol * symbol)
{
  if (lookup (symbol->get_name ()) != NULL)
    throw std::string ("SymbolTable::add -- duplicate table entry");

  map_.insert (SymbolTableMapPair (symbol->get_name (), symbol));
}

const Symbol *
SymbolTable::lookup (const std::string & name) const
{
  std::string mangled = Mangler::mangle (name);
  SymbolTableMapIterator iter = map_.find (mangled);
  if (iter == map_.end ())
    return NULL;

  return iter->second;
}

// Pretty-print to create the symbol table listing.
namespace {
  void
  print_out (SymbolTable::SymbolTableMapPair entry,
             std::ostream * outs)
  {
    *outs << entry.second->create_list_entry ();
  }
}
void
SymbolTable::create_listing (std::ostream & outs) const
{
  outs << "Symbol table:" << std::endl;
  if (map_.empty ())
    {
      outs << " (empty)" << std::endl;
      return;
    }

  outs << Symbol::create_list_header ();
  std::for_each (map_.begin (), map_.end (),
                 std::bind2nd (std::ptr_fun (print_out), &outs));
}
